import functools
from collections import OrderedDict
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List

from django.contrib.auth.password_validation import validate_password
from django.core.exceptions import ValidationError as DjangoValidationError
from django.utils.translation import gettext as _
from rest_framework.exceptions import ErrorDetail, ValidationError
from rest_framework.settings import api_settings

from rest_registration.utils.users import (
    build_initial_user,
    get_user_by_verification_id
)

if TYPE_CHECKING:
    from django.contrib.auth.base_user import AbstractBaseUser

Validator = Callable[[Any], None]


def wrap_validation_error_with_field(
        field_name: str) -> Callable[[Validator], Validator]:

    def decorator(func: Validator) -> Validator:

        @functools.wraps(func)
        def wrapper(value: Any) -> None:
            try:
                func(value)
            except ValidationError as exc:
                raise ValidationError({field_name: exc.detail}) from None

        return wrapper

    return decorator


@wrap_validation_error_with_field('password_confirm')
def validate_user_password_confirm(user_data: Dict[str, Any]) -> None:
    if user_data['password'] != user_data['password_confirm']:
        raise ValidationError(ErrorDetail(
            _("Passwords don't match"),
            code='passwords-do-not-match'),
        )


@wrap_validation_error_with_field('password')
def validate_user_password(user_data: Dict[str, Any]) -> None:
    password = user_data['password']
    user = build_initial_user(user_data)
    return _validate_user_password(password, user)


@wrap_validation_error_with_field('password')
def validate_password_with_user_id(user_data: Dict[str, Any]) -> None:
    password = user_data['password']
    user_id = user_data['user_id']
    user = get_user_by_verification_id(user_id, require_verified=False)
    return _validate_user_password(password, user)


def _validate_user_password(password: str, user: 'AbstractBaseUser') -> None:
    try:
        validate_password(password, user=user)
    except DjangoValidationError as exc:
        raise ValidationError(list(exc.messages)) from None


def run_validators(validators: Iterable[Validator], value: Any) -> None:
    fields_errors = OrderedDict()  # type: Dict[str, Any]
    non_field_errors = []  # type: List[Any]
    for validator in validators:
        try:
            validator(value)
        except ValidationError as exc:
            if isinstance(exc.detail, Mapping):
                for field_name, field_errors in exc.detail.items():
                    fields_errors.setdefault(field_name, []).extend(
                        field_errors)
            elif isinstance(exc.detail, list):
                non_field_errors.extend(exc.detail)

    if fields_errors:
        errors = {}
        errors.update(fields_errors)
        errors.setdefault(
            api_settings.NON_FIELD_ERRORS_KEY, []).extend(non_field_errors)
        raise ValidationError(errors)
    if non_field_errors:
        raise ValidationError(non_field_errors)



from django.test.signals import setting_changed

from rest_registration.utils.nested_settings import NestedSettings

DEFAULTS = {
    'VERIFY_REGISTRATION_SUCCESS_URL': None,
    'VERIFY_REGISTRATION_FAILURE_URL': None,

    'VERIFY_EMAIL_SUCCESS_URL': None,
    'VERIFY_EMAIL_FAILURE_URL': None,

    'RESET_PASSWORD_SUCCESS_URL': None,
    'RESET_PASSWORD_FAILURE_URL': None,
}

IMPORT_STRINGS = ()

verification_redirects_settings = NestedSettings(
    None, DEFAULTS, IMPORT_STRINGS,
    root_setting_name='REST_REGISTRATION_VERIFICATION_REDIRECTS')


def settings_changed_handler(*args, **kwargs) -> None:
    verification_redirects_settings.reset_user_settings()
    verification_redirects_settings.reset_attr_cache()


setting_changed.connect(settings_changed_handler)



from collections import namedtuple
from typing import TYPE_CHECKING, Any, Dict

from django.core.exceptions import ImproperlyConfigured
from django.core.mail.message import EmailMultiAlternatives
from django.template.exceptions import TemplateDoesNotExist
from django.template.loader import get_template, render_to_string
from django.utils.translation import gettext as _

from rest_registration.notifications.enums import NotificationMethod, NotificationType
from rest_registration.settings import registration_settings
from rest_registration.utils.common import identity
from rest_registration.utils.users import get_user_email_field_name

if TYPE_CHECKING:
    from django.contrib.auth.base_user import AbstractBaseUser

EmailTemplateConfig = namedtuple('EmailTemplateConfig', (
    'subject_template_name',
    'text_body_template_name',
    'html_body_template_name',
    'text_body_processor',
))


def send_verification_notification(
        notification_type: NotificationType,
        user: 'AbstractBaseUser',
        data: Dict[str, Any],
        template_config_data: Dict[str, Any],
        custom_user_address: Any = None) -> None:
    if custom_user_address is None:
        user_address = get_user_address(user)
    else:
        user_address = custom_user_address
    notification = create_verification_notification(
        notification_type, user, user_address, data, template_config_data)
    send_notification(notification)


def create_verification_notification(
        notification_type: NotificationType,
        user: 'AbstractBaseUser',
        user_address: Any,
        data: Dict[str, Any],
        template_config_data: Dict[str, Any]) -> EmailMultiAlternatives:
    from_email = registration_settings.VERIFICATION_FROM_EMAIL
    reply_to_email = (registration_settings.VERIFICATION_REPLY_TO_EMAIL or
                      from_email)
    template_context_builder = registration_settings.VERIFICATION_TEMPLATE_CONTEXT_BUILDER  # noqa: E501
    context = template_context_builder(
        user, user_address, data,
        notification_type=notification_type,
        notification_method=NotificationMethod.EMAIL)
    template_config = parse_template_config(template_config_data)

    subject = render_to_string(
        template_config.subject_template_name, context=context).strip()
    text_body = template_config.text_body_processor(
        render_to_string(
            template_config.text_body_template_name, context=context))

    email_msg = EmailMultiAlternatives(
        subject, text_body, from_email, [user_address],
        reply_to=[reply_to_email])

    if template_config.html_body_template_name:
        html_body = render_to_string(
            template_config.html_body_template_name, context=context)
        email_msg.attach_alternative(html_body, 'text/html')

    return email_msg


def send_notification(notification: EmailMultiAlternatives) -> None:
    notification.send()


def get_user_address(user: 'AbstractBaseUser') -> str:
    email_field_name = get_user_email_field_name()
    email = getattr(user, email_field_name)  # type: str
    return email


def parse_template_config(template_config_data: Dict[str, Any]) -> EmailTemplateConfig:
    """
    >>> from tests import doctest_utils
    >>> convert_html_to_text = registration_settings.VERIFICATION_EMAIL_HTML_TO_TEXT_CONVERTER  # noqa: E501
    >>> parse_template_config({})  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ImproperlyConfigured
    >>> parse_template_config({
    ...     'subject': 'blah',
    ... })  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ImproperlyConfigured
    >>> parse_template_config({
    ...     'subject': 'blah',
    ...     'body': 'blah',
    ... })  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ImproperlyConfigured
    >>> doctest_utils.equals(
    ...     parse_template_config({
    ...         'subject': 'rest_registration/register/subject.txt',
    ...         'html_body': 'rest_registration/register/body.html',
    ...         'text_body': 'rest_registration/register/body.txt',
    ...     }),
    ...     EmailTemplateConfig(
    ...         'rest_registration/register/subject.txt',
    ...         'rest_registration/register/body.txt',
    ...         'rest_registration/register/body.html',
    ...         identity))
    OK
    >>> doctest_utils.equals(
    ...     parse_template_config({
    ...         'subject': 'rest_registration/register/subject.txt',
    ...         'html_body': 'rest_registration/register/body.html',
    ...     }),
    ...     EmailTemplateConfig(
    ...         'rest_registration/register/subject.txt',
    ...         'rest_registration/register/body.html',
    ...         'rest_registration/register/body.html',
    ...         convert_html_to_text))
    OK
    >>> doctest_utils.equals(
    ...     parse_template_config({
    ...         'subject': 'rest_registration/register/subject.txt',
    ...         'text_body': 'rest_registration/register/body.txt',
    ...     }),
    ...     EmailTemplateConfig(
    ...         'rest_registration/register/subject.txt',
    ...         'rest_registration/register/body.txt', None,
    ...         identity))
    OK
    >>> doctest_utils.equals(
    ...     parse_template_config({
    ...         'subject': 'rest_registration/register/subject.txt',
    ...         'body': 'rest_registration/register/body.txt',
    ...     }),
    ...     EmailTemplateConfig(
    ...         'rest_registration/register/subject.txt',
    ...         'rest_registration/register/body.txt', None,
    ...         identity))
    OK
    >>> doctest_utils.equals(
    ...     parse_template_config({
    ...         'subject': 'rest_registration/register/subject.txt',
    ...         'body': 'rest_registration/register/body.html',
    ...         'is_html': True,
    ...     }),
    ...     EmailTemplateConfig(
    ...         'rest_registration/register/subject.txt',
    ...         'rest_registration/register/body.html',
    ...         'rest_registration/register/body.html',
    ...         convert_html_to_text))
    OK
    """
    try:
        subject_template_name = template_config_data['subject']
    except KeyError:
        raise ImproperlyConfigured(_("No 'subject' key found")) from None
    body_template_name = template_config_data.get('body')
    text_body_template_name = template_config_data.get('text_body')
    html_body_template_name = template_config_data.get('html_body')
    is_html_body = template_config_data.get('is_html')
    convert_html_to_text = registration_settings.VERIFICATION_EMAIL_HTML_TO_TEXT_CONVERTER  # noqa: E501

    if html_body_template_name and text_body_template_name:
        config = EmailTemplateConfig(
            subject_template_name=subject_template_name,
            text_body_template_name=text_body_template_name,
            html_body_template_name=html_body_template_name,
            text_body_processor=identity,
        )
    elif html_body_template_name:
        config = EmailTemplateConfig(
            subject_template_name=subject_template_name,
            text_body_template_name=html_body_template_name,
            html_body_template_name=html_body_template_name,
            text_body_processor=convert_html_to_text,
        )
    elif text_body_template_name:
        config = EmailTemplateConfig(
            subject_template_name=subject_template_name,
            text_body_template_name=text_body_template_name,
            html_body_template_name=None,
            text_body_processor=identity,
        )
    elif body_template_name:
        if is_html_body:
            config = EmailTemplateConfig(
                subject_template_name=subject_template_name,
                text_body_template_name=body_template_name,
                html_body_template_name=body_template_name,
                text_body_processor=convert_html_to_text,
            )
        else:
            config = EmailTemplateConfig(
                subject_template_name=subject_template_name,
                text_body_template_name=body_template_name,
                html_body_template_name=None,
                text_body_processor=identity,
            )
    else:
        raise ImproperlyConfigured(
            'Could not parse template config data: {template_config_data}'.format(  # noqa: E501
                template_config_data=template_config_data))

    _validate_template_name_existence(config.subject_template_name)
    _validate_template_name_existence(config.text_body_template_name)

    if config.html_body_template_name:
        _validate_template_name_existence(config.html_body_template_name)

    assert callable(config.text_body_processor)

    return config


def _validate_template_name_existence(template_name: str) -> None:
    try:
        get_template(template_name)
    except TemplateDoesNotExist:
        raise ImproperlyConfigured(
            'Template {template_name!r} does not exist; ensure that your'
            ' Django TEMPLATES setting is configured correctly'.format(
                template_name=template_name)) from None



from typing import TYPE_CHECKING, Any, Dict

from django.contrib import auth
from django.utils.translation import gettext as _
from rest_framework import serializers
from rest_framework.authentication import SessionAuthentication, TokenAuthentication
from rest_framework.decorators import api_view, permission_classes
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.settings import api_settings

from rest_registration.decorators import (
    api_view_serializer_class,
    api_view_serializer_class_getter
)
from rest_registration.exceptions import LoginInvalid, UserNotFound
from rest_registration.settings import registration_settings
from rest_registration.utils.responses import get_ok_response

if TYPE_CHECKING:
    from django.contrib.auth.base_user import AbstractBaseUser


@api_view_serializer_class_getter(
    lambda: registration_settings.LOGIN_SERIALIZER_CLASS)
@api_view(['POST'])
@permission_classes(registration_settings.NOT_AUTHENTICATED_PERMISSION_CLASSES)
def login(request: Request) -> Response:
    '''
    Logs in the user via given login and password.
    '''
    serializer_class = registration_settings.LOGIN_SERIALIZER_CLASS
    serializer = serializer_class(data=request.data, context={'request': request})
    serializer.is_valid(raise_exception=True)
    login_authenticator = registration_settings.LOGIN_AUTHENTICATOR
    try:
        user = login_authenticator(serializer.validated_data, serializer=serializer)
    except UserNotFound:
        raise LoginInvalid() from None

    extra_data = perform_login(request, user)

    return get_ok_response(_("Login successful"), extra_data=extra_data)


class LogoutSerializer(serializers.Serializer):  # pylint: disable=abstract-method
    revoke_token = serializers.BooleanField(default=False)


@api_view_serializer_class(LogoutSerializer)
@api_view(['POST'])
@permission_classes([IsAuthenticated])
def logout(request: Request) -> Response:
    '''
    Logs out the user. returns an error if the user is not
    authenticated.
    '''
    user = request.user
    serializer = LogoutSerializer(
        data=request.data,
        context={'request': request},
    )
    serializer.is_valid(raise_exception=True)
    data = serializer.validated_data

    if should_authenticate_session():
        auth.logout(request)
    if should_retrieve_token() and data['revoke_token']:
        auth_token_manager_cls = registration_settings.AUTH_TOKEN_MANAGER_CLASS
        auth_token_manager = auth_token_manager_cls()  # noqa: E501 type: rest_registration.auth_token_managers.AbstractAuthTokenManager
        auth_token_manager.revoke_token(user)

    return get_ok_response(_("Logout successful"))


def should_authenticate_session() -> bool:
    result = registration_settings.LOGIN_AUTHENTICATE_SESSION
    if result is None:
        result = rest_auth_has_class(SessionAuthentication)
    return result


def should_retrieve_token() -> bool:
    result = registration_settings.LOGIN_RETRIEVE_TOKEN
    if result is None:
        result = rest_auth_has_class(TokenAuthentication)
    return result


def rest_auth_has_class(cls: type) -> bool:
    return cls in api_settings.DEFAULT_AUTHENTICATION_CLASSES


def perform_login(request: Request, user: 'AbstractBaseUser') -> Dict[str, Any]:
    if should_authenticate_session():
        auth.login(request, user)

    extra_data = {}

    if should_retrieve_token():
        auth_token_manager_cls = registration_settings.AUTH_TOKEN_MANAGER_CLASS
        auth_token_manager = auth_token_manager_cls()  # noqa: E501 type: rest_registration.auth_token_managers.AbstractAuthTokenManager
        token = auth_token_manager.provide_token(user)
        extra_data['token'] = token

    return extra_data



from collections import deque
from html.parser import HTMLParser
from typing import Dict, List, Optional, Tuple


def convert_html_to_text_preserving_urls(value: str) -> str:
    return convert_html_to_text(value, preserve_urls=True)


def convert_html_to_text(value: str, preserve_urls: bool = False) -> str:
    stripper = MLStripper(preserve_urls=preserve_urls)
    stripper.feed(value)
    stripper.close()
    return stripper.get_data()


class MLStripper(HTMLParser):

    def __init__(self, preserve_urls: bool = False) -> None:
        super().__init__(convert_charrefs=True)
        self.reset()
        self._paragraphs = [[]]  # type: List[List[str]]
        self._tag_info_stack = deque([TagInfo(None, {})])
        self._preserve_urls = preserve_urls

    def handle_starttag(self, tag: str, attrs: List[Tuple[str, Optional[str]]]) -> None:
        self._tag_info_stack.append(TagInfo(tag, dict(attrs)))
        if tag == 'br':
            self._paragraphs.append([])

    def handle_endtag(self, tag: str) -> None:
        tag_info = self._tag_info_stack[-1]
        href = tag_info.attrs.get('href')
        last_paragraph = self._paragraphs[-1]
        last_segment = last_paragraph[-1] if last_paragraph else None
        if (self._preserve_urls and
                tag_info.name == 'a' and href and href != last_segment):
            self._append_segment('({href})'.format(href=href))

        if tag == 'p':
            self._paragraphs.append([])

        self._tag_info_stack.pop()

    def handle_data(self, data: str) -> None:
        data = ' '.join(data.split())
        if not self._is_in_body():
            return
        if data:
            self._append_segment(data)

    def error(self, message: str) -> None:
        raise ValueError("HTML parse error: {message}".format(
            message=message))

    def get_data(self) -> str:
        paragraph_texts = []
        for segments in self._paragraphs:
            paragraph_texts.append(' '.join(segments))
        return '\n'.join(paragraph_texts)

    def _is_in_body(self) -> bool:
        if len(self._tag_info_stack) < 3:
            return False
        return (self._tag_info_stack[1].name == 'html' and
                self._tag_info_stack[2].name == 'body')

    def _append_segment(self, segment: str) -> None:
        self._paragraphs[-1].append(segment)


class TagInfo:

    def __init__(self, name: Optional[str], attrs: Dict[str, Optional[str]]) -> None:
        super().__init__()
        self._name = name
        self._attrs = attrs

    @property
    def name(self) -> Optional[str]:
        return self._name

    @property
    def attrs(self) -> Dict[str, Optional[str]]:
        return self._attrs



from functools import WRAPPER_ASSIGNMENTS, WRAPPER_UPDATES
from typing import Any, Callable, Iterable, TypeVar

_T = TypeVar('_T')


# Copied from stdlib functools module
def update_wrapper(
        wrapper: Any,
        wrapped: Callable[..., _T],
        assigned: Iterable[str] = WRAPPER_ASSIGNMENTS,
        updated: Iterable[str] = WRAPPER_UPDATES,
        set_wrapped: bool = True):
    for attr in assigned:
        try:
            value = getattr(wrapped, attr)
        except AttributeError:
            pass
        else:
            setattr(wrapper, attr, value)
    for attr in updated:
        getattr(wrapper, attr).update(getattr(wrapped, attr, {}))
    if set_wrapped:
        # Issue #17482: set __wrapped__ last so we don't inadvertently copy it
        # from the wrapped function when updating __dict__
        wrapper.__wrapped__ = wrapped
    # Return the wrapper so this can be used as a decorator via partial()
    return wrapper



from django.conf.urls import url

from .views import reset_password, verify_email, verify_registration

app_name = 'rest_registration.contrib.verification_redirects'
urlpatterns = [
    url('^verify-registration/$', verify_registration,
        name='verify-registration'),
    url('^verify-email/$', verify_email, name='verify-email'),
    url('^reset-password/$', reset_password, name='reset-password'),
]



from typing import TYPE_CHECKING, Dict

from rest_framework.request import Request

from rest_registration.exceptions import VerificationTemplatesNotFound
from rest_registration.notifications.email import send_verification_notification
from rest_registration.notifications.enums import NotificationMethod, NotificationType
from rest_registration.settings import registration_settings
from rest_registration.signers.register import RegisterSigner
from rest_registration.signers.register_email import RegisterEmailSigner
from rest_registration.signers.reset_password import ResetPasswordSigner
from rest_registration.utils.users import get_user_verification_id
from rest_registration.utils.verification import select_default_templates

if TYPE_CHECKING:
    from django.contrib.auth.base_user import AbstractBaseUser


def send_register_verification_email_notification(
        request: Request,
        user: 'AbstractBaseUser',
) -> None:
    signer = RegisterSigner({
        'user_id': get_user_verification_id(user),
    }, request=request)
    template_config_data = _get_email_template_config_data(
        request, user, NotificationType.REGISTER_VERIFICATION)
    notification_data = {
        'params_signer': signer,
    }
    send_verification_notification(
        NotificationType.REGISTER_VERIFICATION, user,
        notification_data, template_config_data)


def send_register_email_verification_email_notification(
        request: Request,
        user: 'AbstractBaseUser',
        email: str,
        email_already_used: bool = False,
) -> None:
    signer = RegisterEmailSigner({
        'user_id': get_user_verification_id(user),
        'email': email,
    }, request=request)
    notification_data = {
        'params_signer': signer,
        'email_already_used': email_already_used,
    }
    template_config_data = _get_email_template_config_data(
        request, user, NotificationType.REGISTER_EMAIL_VERIFICATION)
    send_verification_notification(
        NotificationType.REGISTER_EMAIL_VERIFICATION, user,
        notification_data, template_config_data, custom_user_address=email)


def send_reset_password_verification_email_notification(
        request: Request,
        user: 'AbstractBaseUser',
) -> None:
    signer = ResetPasswordSigner({
        'user_id': get_user_verification_id(user),
    }, request=request)

    template_config_data = _get_email_template_config_data(
        request, user, NotificationType.RESET_PASSWORD_VERIFICATION)
    notification_data = {
        'params_signer': signer,
    }
    send_verification_notification(
        NotificationType.RESET_PASSWORD_VERIFICATION, user, notification_data,
        template_config_data)


def _get_email_template_config_data(
        request: Request,
        user: 'AbstractBaseUser',
        notification_type: NotificationType,
) -> Dict[str, str]:
    template_selector = registration_settings.VERIFICATION_TEMPLATES_SELECTOR
    notification_method = NotificationMethod.EMAIL
    try:
        template_config_data = template_selector(
            request=request,
            user=user,
            notification_method=notification_method,
            notification_type=notification_type,
        )
    except (VerificationTemplatesNotFound, LookupError):
        template_config_data = select_default_templates(
            request=request,
            user=user,
            notification_method=notification_method,
            notification_type=notification_type,
        )
    return template_config_data



from django.db import transaction
from django.http import Http404
from django.utils.translation import gettext as _
from rest_framework import serializers, status
from rest_framework.decorators import api_view, permission_classes
from rest_framework.response import Response

from rest_registration import signals
from rest_registration.api.views.login import perform_login
from rest_registration.decorators import (
    api_view_serializer_class,
    api_view_serializer_class_getter
)
from rest_registration.exceptions import UserWithoutEmailNonverifiable
from rest_registration.settings import registration_settings
from rest_registration.signers.register import RegisterSigner
from rest_registration.utils.responses import get_ok_response
from rest_registration.utils.users import (
    get_user_by_verification_id,
    get_user_email_field_name,
    get_user_setting
)
from rest_registration.utils.verification import verify_signer_or_bad_request
from rest_registration.verification_notifications import (
    send_register_verification_email_notification
)


@api_view_serializer_class_getter(
    lambda: registration_settings.REGISTER_SERIALIZER_CLASS)
@api_view(['POST'])
@permission_classes(registration_settings.NOT_AUTHENTICATED_PERMISSION_CLASSES)
def register(request):
    '''
    Register new user.
    '''
    serializer_class = registration_settings.REGISTER_SERIALIZER_CLASS
    serializer = serializer_class(data=request.data, context={'request': request})
    serializer.is_valid(raise_exception=True)

    kwargs = {}

    if registration_settings.REGISTER_VERIFICATION_ENABLED:
        verification_flag_field = get_user_setting('VERIFICATION_FLAG_FIELD')
        kwargs[verification_flag_field] = False
        email_field_name = get_user_email_field_name()
        if (email_field_name not in serializer.validated_data
                or not serializer.validated_data[email_field_name]):
            raise UserWithoutEmailNonverifiable()

    with transaction.atomic():
        user = serializer.save(**kwargs)
        if registration_settings.REGISTER_VERIFICATION_ENABLED:
            send_register_verification_email_notification(request, user)

    signals.user_registered.send(sender=None, user=user, request=request)
    output_serializer_class = registration_settings.REGISTER_OUTPUT_SERIALIZER_CLASS
    output_serializer = output_serializer_class(
        instance=user,
        context={'request': request},
    )
    user_data = output_serializer.data
    return Response(user_data, status=status.HTTP_201_CREATED)


class VerifyRegistrationSerializer(serializers.Serializer):  # noqa: E501 pylint: disable=abstract-method
    user_id = serializers.CharField(required=True)
    timestamp = serializers.IntegerField(required=True)
    signature = serializers.CharField(required=True)


@api_view_serializer_class(VerifyRegistrationSerializer)
@api_view(['POST'])
@permission_classes(registration_settings.NOT_AUTHENTICATED_PERMISSION_CLASSES)
def verify_registration(request):
    """
    Verify registration via signature.
    """
    user = process_verify_registration_data(
        request.data, serializer_context={'request': request})
    signals.user_activated.send(sender=None, user=user, request=request)
    extra_data = None
    if registration_settings.REGISTER_VERIFICATION_AUTO_LOGIN:
        extra_data = perform_login(request, user)
    return get_ok_response(_("User verified successfully"), extra_data=extra_data)


def process_verify_registration_data(input_data, serializer_context=None):
    if serializer_context is None:
        serializer_context = {}
    if not registration_settings.REGISTER_VERIFICATION_ENABLED:
        raise Http404()
    serializer = VerifyRegistrationSerializer(
        data=input_data,
        context=serializer_context,
    )
    serializer.is_valid(raise_exception=True)

    data = serializer.validated_data
    signer = RegisterSigner(data)
    verify_signer_or_bad_request(signer)

    verification_flag_field = get_user_setting('VERIFICATION_FLAG_FIELD')
    user = get_user_by_verification_id(data['user_id'], require_verified=False)
    setattr(user, verification_flag_field, True)
    user.save()

    return user



from typing import Any, Dict, Optional

from django.http import Http404
from django.utils.translation import gettext as _
from rest_framework import serializers
from rest_framework.decorators import api_view, permission_classes
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response

from rest_registration import signals
from rest_registration.decorators import (
    api_view_serializer_class,
    api_view_serializer_class_getter
)
from rest_registration.exceptions import EmailAlreadyRegistered
from rest_registration.settings import registration_settings
from rest_registration.signers.register_email import RegisterEmailSigner
from rest_registration.utils.responses import get_ok_response
from rest_registration.utils.users import (
    get_user_by_verification_id,
    get_user_email_field_name,
    is_user_email_field_unique,
    user_with_email_exists
)
from rest_registration.utils.verification import verify_signer_or_bad_request
from rest_registration.verification_notifications import (
    send_register_email_verification_email_notification
)


@api_view_serializer_class_getter(
    lambda: registration_settings.REGISTER_EMAIL_SERIALIZER_CLASS)
@api_view(['POST'])
@permission_classes([IsAuthenticated])
def register_email(request: Request) -> Response:
    '''
    Register new email.
    '''
    user = request.user

    serializer_class = registration_settings.REGISTER_EMAIL_SERIALIZER_CLASS
    serializer = serializer_class(data=request.data, context={'request': request})
    serializer.is_valid(raise_exception=True)

    # TODO: Issue #114 - remove code supporting deprecated behavior
    get_email = getattr(serializer, 'get_email', None)
    if callable(get_email):
        email = get_email()
    else:
        email = serializer.validated_data['email']

    email_already_used = is_user_email_field_unique() and user_with_email_exists(email)

    if registration_settings.REGISTER_EMAIL_VERIFICATION_ENABLED:
        send_register_email_verification_email_notification(
            request, user, email, email_already_used=email_already_used)
    else:
        if email_already_used:
            raise EmailAlreadyRegistered()

        email_field_name = get_user_email_field_name()
        old_email = getattr(user, email_field_name)
        setattr(user, email_field_name, email)
        user.save()
        signals.user_changed_email.send(
            sender=None,
            user=user,
            new_email=email,
            old_email=old_email,
            request=request,
        )

    return get_ok_response(_("Register email link email sent"))


class VerifyEmailSerializer(serializers.Serializer):  # noqa: E501 pylint: disable=abstract-method
    user_id = serializers.CharField(required=True)
    email = serializers.EmailField(required=True)
    timestamp = serializers.IntegerField(required=True)
    signature = serializers.CharField(required=True)


@api_view_serializer_class(VerifyEmailSerializer)
@api_view(['POST'])
@permission_classes(registration_settings.NOT_AUTHENTICATED_PERMISSION_CLASSES)
def verify_email(request: Request) -> Response:
    '''
    Verify email via signature.
    '''
    process_verify_email_data(request.data, serializer_context={'request': request})
    return get_ok_response(_("Email verified successfully"))


def process_verify_email_data(
        input_data: Dict[str, Any],
        serializer_context: Optional[Dict[str, Any]] = None) -> None:
    if serializer_context is None:
        serializer_context = {}
    if not registration_settings.REGISTER_EMAIL_VERIFICATION_ENABLED:
        raise Http404()
    serializer = VerifyEmailSerializer(data=input_data, context=serializer_context)
    serializer.is_valid(raise_exception=True)

    data = serializer.validated_data
    signer = RegisterEmailSigner(data)
    verify_signer_or_bad_request(signer)
    request = serializer_context.get('request')
    new_email = data['email']

    if is_user_email_field_unique() and user_with_email_exists(new_email):
        raise EmailAlreadyRegistered()

    email_field_name = get_user_email_field_name()
    user = get_user_by_verification_id(data['user_id'])
    old_email = getattr(user, email_field_name)
    setattr(user, email_field_name, new_email)
    user.save()

    signals.user_changed_email.send(
        sender=None,
        user=user,
        new_email=new_email,
        old_email=old_email,
        request=request,
    )



from rest_registration.settings import registration_settings
from rest_registration.utils.signers import URLParamsSigner
from rest_registration.utils.users import get_user_by_verification_id


class ResetPasswordSigner(URLParamsSigner):
    SALT_BASE = 'reset-password'
    USE_TIMESTAMP = True

    def get_base_url(self):
        return registration_settings.RESET_PASSWORD_VERIFICATION_URL

    def get_valid_period(self):
        return registration_settings.RESET_PASSWORD_VERIFICATION_PERIOD

    def _calculate_salt(self, data):
        if registration_settings.RESET_PASSWORD_VERIFICATION_ONE_TIME_USE:
            user = get_user_by_verification_id(data['user_id'], require_verified=False)
            user_password_hash = user.password
            # Use current user password hash as a part of the salt.
            # If the password gets changed, then assume that the change
            # was caused by previous password reset and the signature
            # is not valid anymore because changed password hash implies
            # changed salt used when verifying the input data.
            salt = '{self.SALT_BASE}:{user_password_hash}'.format(
                self=self, user_password_hash=user_password_hash)
        else:
            salt = self.SALT_BASE
        return salt



import functools
from typing import Any, Callable, Iterable, List, Type, Union

from django.apps import AppConfig
from django.core import checks

from rest_registration.utils.functools import update_wrapper

WRAPPER_ASSIGNMENTS_WITHOUT_ANNOTATIONS = tuple(
    a for a in functools.WRAPPER_ASSIGNMENTS if a != '__annotations__')

CheckMessage = Union[checks.Error, checks.Warning]


class CheckCode:

    def get_full_code_id(self) -> str:
        app_name = self.get_app_name()
        code_id = self.get_code_id()
        return '{app_name}.{code_id}'.format(app_name=app_name, code_id=code_id)

    def get_app_name(self) -> str:
        raise NotImplementedError()

    def get_code_id(self) -> str:
        raise NotImplementedError()

    def get_check_message_class(self) -> Type[CheckMessage]:
        raise NotImplementedError()


def predicate_check(
        error_message: str, error_code: CheckCode,
        obj: Any = None
) -> Callable[[Callable[[], bool]], Callable[..., List[checks.CheckMessage]]]:
    message_cls = error_code.get_check_message_class()

    def decorator(
            predicate: Callable[[], bool]
    ) -> Callable[..., List[checks.CheckMessage]]:

        def check_fun(
                app_configs: Iterable[AppConfig], **kwargs,
        ) -> List[checks.CheckMessage]:
            if predicate():
                return []

            return [
                message_cls(
                    error_message,
                    obj=obj,
                    hint=None,
                    id=error_code.get_full_code_id(),
                )
            ]

        update_wrapper(
            check_fun, predicate,
            assigned=WRAPPER_ASSIGNMENTS_WITHOUT_ANNOTATIONS,
            set_wrapped=False)

        return check_fun

    return decorator


def no_exception_check(
        error_message: str, error_code: CheckCode,
        obj: Any = None
) -> Callable[[Callable[[], None]], Callable[..., List[checks.CheckMessage]]]:
    message_cls = error_code.get_check_message_class()

    def decorator(
            fun: Callable[[], None]
    ) -> Callable[..., List[checks.CheckMessage]]:

        def check_fun(
                app_configs: Iterable[AppConfig], **kwargs,
        ) -> List[checks.CheckMessage]:
            try:
                fun()
            except Exception as exc:  # pylint: disable=broad-except
                exc_str = str(exc)
                msg = '{error_message}: {exc_str}'.format(
                    error_message=error_message, exc_str=exc_str)
                return [
                    message_cls(
                        msg,
                        obj=obj,
                        hint=None,
                        id=error_code.get_full_code_id(),
                    )
                ]
            else:
                return []

        update_wrapper(
            check_fun, fun,
            assigned=WRAPPER_ASSIGNMENTS_WITHOUT_ANNOTATIONS,
            set_wrapped=False)

        return check_fun

    return decorator



from typing import TYPE_CHECKING, NewType, Optional, Sequence, Type

from rest_framework.authentication import BaseAuthentication, TokenAuthentication

from rest_registration.exceptions import AuthTokenNotFound, AuthTokenNotRevoked

AuthToken = NewType('AuthToken', str)

if TYPE_CHECKING:
    from django.contrib.auth.base_user import AbstractBaseUser


class AbstractAuthTokenManager:

    def get_authentication_class(self) -> Type[BaseAuthentication]:
        """
        Return authentication class which is able to parse the token.
        This is used to ensure that the class is added
        in ``REST_FRAMEWORK['DEFAULT_AUTHENTICATION_CLASSES']`` setting.
        """
        raise NotImplementedError()

    def get_app_names(self) -> Sequence[str]:
        """
        Return the Django app names which need to be installed so
        this token manager class works properly.

        Overriding this method is not required but recommended as it provides
        additional check during the Django startup.
        """
        return []

    def provide_token(self, user: 'AbstractBaseUser') -> AuthToken:
        """
        Get or create token for given user.
        If there is no token to provide,
        raise ``rest_registration.exceptions.AuthTokenError``.
        """
        raise NotImplementedError()

    def revoke_token(
            self, user: 'AbstractBaseUser', *,
            token: Optional[AuthToken] = None) -> None:
        """
        Revoke the given token for a given user. If the token is not provided,
        revoke all tokens for given user.
        If the provided token is invalid or there is no token to revoke,
        raise ``rest_registration.exceptions.AuthTokenError``.

        This method may not be implemented in all cases - for instance, in case
        when the token is cryptographically generated and not stored
        in the database.
        """
        raise AuthTokenNotRevoked()


class RestFrameworkAuthTokenManager(AbstractAuthTokenManager):

    def get_authentication_class(self) -> Type[BaseAuthentication]:
        return TokenAuthentication

    def get_app_names(self) -> Sequence[str]:
        return [
            'rest_framework.authtoken',
        ]

    def provide_token(self, user: 'AbstractBaseUser') -> AuthToken:
        from rest_framework.authtoken.models import Token  # noqa: E501 pylint: disable=import-outside-toplevel

        token_obj, _ = Token.objects.get_or_create(user=user)
        return AuthToken(token_obj.key)

    def revoke_token(
            self, user: 'AbstractBaseUser', *,
            token: Optional[AuthToken] = None) -> None:
        from rest_framework.authtoken.models import Token  # noqa: E501 pylint: disable=import-outside-toplevel

        try:
            token_obj = Token.objects.get(user_id=user.pk)  # type: Token
        except Token.DoesNotExist:
            raise AuthTokenNotFound() from None

        if token is not None and token_obj.key != token:
            raise AuthTokenNotFound() from None

        token_obj.delete()



from typing import Callable, Iterable, Optional, Set, TypeVar, Union


class RaiseExceptionType:
    pass


_T = TypeVar('_T')
LazyBool = Callable[[], bool]
MaybeLazyBool = Union[bool, LazyBool]


RAISE_EXCEPTION = RaiseExceptionType()


def identity(value: _T) -> _T:
    """
    >>> identity(None)
    >>> identity(1)
    1
    """
    return value


def materialize_bool(value: MaybeLazyBool) -> bool:
    """
    >>> materialize_bool(True)
    True
    >>> materialize_bool(False)
    False
    >>> materialize_bool(lambda: True)
    True
    >>> materialize_bool(lambda: False)
    False
    """
    return value() if callable(value) else value


def implies(premise: bool, conclusion: MaybeLazyBool) -> bool:
    """
    Calculate material implication for given premise and conclusion.
    The conclusion may be lazy evaluated if it is of LazyBool type.
    >>> implies(True, True)
    True
    >>> implies(True, False)
    False
    >>> implies(False, True)
    True
    >>> implies(False, False)
    True
    >>> implies(True, lambda: True)
    True
    >>> implies(True, lambda: False)
    False
    >>> implies(False, lambda: True)
    True
    >>> implies(False, lambda: False)
    True
    """
    if not premise:
        return True
    return materialize_bool(conclusion)


def set_or_none(iterable: Optional[Iterable[_T]]) -> Optional[Set[_T]]:
    """
    >>> set_or_none(None)
    >>> set_or_none([1,1,2]) == {1,2}
    True
    """
    if iterable is None:
        return None
    return set(iterable)
